import tokenize, io, typer, typing, os, sys, pathlib, traceback

try:
    import python_minifier as minifier
except ImportError as e:
    traceback.print_exception(e)
    minifier_avail = False
else:
    minifier_avail = True

from std.ipathlib import Path
def patch_std(prefix: str):
    for i in (Path(__file__).parent.parent / 'std').iterdir():
        if i.is_dir(): continue
        module = i.read_text()
        prefix = prefix.replace(f"std'{i.name}'", "'''" + module.replace("'''", "\\'''") + "'''")
    
    
    return prefix

PREFIX = patch_std((Path(__file__).parent / 'prefix.py').read_text())

class lazy_typegetter:
    class lazy_typegetter_iter:
        def __init__(self, typegetter) -> None:
            self.typegetter = typegetter
            self.index = 0
        def __next__(self):
            try:
                val: tuple[int, str] = self.typegetter[self.index]
            except IndexError:
                raise StopIteration
            self.index+=1
            return val
    def __init__(self, tokens: list) -> None:
        self.tokens = tokens
    def __getitem__(self, item: int):
        type, name, _, _, _ = self.tokens[item]
        return type, name

    def __iter__(self):
        return self.lazy_typegetter_iter(self)



def translate(file: io.StringIO, debug: int = 0, bcp: bool = False):
    def infix(name: str):
        yield tokenize.OP, ">>"
        yield tokenize.NAME, name
        yield tokenize.OP, "<<"
    def dprint(data: str, prio: int):
        if debug < prio:
            return
        print(data)
    fdata = file.read()
    filedata = PREFIX + fdata
    dprint(f'---START FILE---\nname: {file.name}, len: {len(fdata)}\n' + filedata + '\n---END FILE---\n', 5)
    patched_file = io.StringIO(filedata)
    
    skip_token = 0
    tokens = lazy_typegetter(list(tokenize.generate_tokens(patched_file.readline))) # Precalculate tokens
    dprint(f'---START DEBOUT---\nname: {file.name}', 4)
    for n, i in enumerate(tokens):
        type, name = i
        dprint(str(type) + " " + name, 4)
        try:
            next_type, next_name = tokens[n + 1]
        except IndexError:
            next_type, next_name = (None, None)
        
        if skip_token:
            skip_token -= 1
            continue
        if type == tokenize.OP and next_type == tokenize.OP:
            dprint(f"OP {name} {next_name} {tokens[n + 2][1]}", 5)
            # Most likely special operation
            if name == "|" and next_name == ">": # left pipe
                yield from infix("_INTERNAL_lpipe")
                skip_token+=1
            elif name == "<" and next_name == "|": # right pipe
                yield from infix("_INTERNAL_rpipe")
                skip_token+=1
            elif name == ":" and next_name == ":": # lazy merge
                yield from infix("_INTERNAL_lazymerge")
                skip_token+=1
            elif name == "+" and next_name == "+": # increment
                yield tokenize.OP, "+="
                yield tokenize.NUMBER, "1"
                skip_token+=1
            elif name == "-" and next_name == "-": # decrement
                yield tokenize.OP, "-="
                yield tokenize.NUMBER, "1"
                skip_token+=1
            elif name == ")" and next_name == "=": # short functions
                yield type, name
                yield tokenize.OP, ":"
                yield tokenize.NAME, "return"

                skip_token += 1
            elif name == "?" and next_name == "?":
                yield from infix("_INTERNAL_nonereplace")
                skip_token+=1
            elif name == "|" and next_name == "*" and tokens[n + 2][1] == ">": #|*>
                yield from infix("_INTERNAL_lspipe")
                skip_token+=2
            elif name == "<" and next_name == "*" and tokens[n + 2][1] == "|": #<*|
                yield from infix("_INTERNAL_rspipe")
                skip_token+=2
            elif name == '/' and next_name == "*":
                yield tokenize.NEWLINE, '\n'
                yield tokenize.OP, "'''"
                skip_token+=1
            elif name == '*' and next_name == "/":
                yield tokenize.OP, "'''"
                skip_token+=1
            elif name == '+' and next_name == ">":
                yield tokenize.NAME, "yield"
                skip_token+=1
            
            elif name == '+' and next_name == ">>":
                yield tokenize.NAME, "yield"
                yield tokenize.NAME, "from"
                skip_token+=2
            else:
                yield type,name

        elif type == tokenize.OP:

            if name == "<>": # barry is flufl
                yield tokenize.OP, "!="
            elif tokens[n - 1][1] != ')' and name == "->": # -> operator, don't trigger on function type annotation
                pass
            else:
                yield type, name
        #elif type == tokenize.NAME and next_name == ':':
        #    # Custom control flow statement
        #    if name == 'local':
        #        yield 
        #    else:
        #        yield type, name
        #elif type == tokenize.NAME:
        #    if name == "λ":
        #        yield tokenize.NAME, 'lambda'
        #    else:
        #        yield type, name
        else:
            yield type,name
    dprint(f'---END DEBOUT---', 4)

def transpile(input_path: Path, verbosity: int, minify: bool, bcp: bool) -> None:
    dir = Path('dist')
    if input_path.is_dir():
        for i in input_path.glob('*'):
            if i.is_dir(): continue
            path = (dir / i.relative_to(input_path)).with_suffix('.py')
            path.parent.mkdir(parents=True, exist_ok=True)
            with i.open() as f:
                transpiled = tokenize.untokenize(translate(f, verbosity))
                if minify:
                    if not minifier_avail:
                        print('ERROR: Minifier not available.')
                        return
                    transpiled = minifier.minify(transpiled)
                path.write_text(transpiled)
    else:
        with input_path.open() as f:
            transpiled = tokenize.untokenize(translate(f, verbosity))
            if minify:
                if not minifier_avail:
                    print('ERROR: Minifier not available.')
                    return
                transpiled = minifier.minify(transpiled)
            input_path.with_suffix('.py').write_text(transpiled)

app = typer.Typer()

verbosity_arg = typing.Annotated[int, typer.Option('--verbosity', '-v')]
minify_arg = typing.Annotated[bool, typer.Option('--minify', '-m')]
bcp_arg = typing.Annotated[bool, typer.Option('--build-custom-prefix', '-B')]


@app.command('t')
@app.command('ts')
@app.command('transpile')
def transpile_cmd(input_path: pathlib.Path, verbosity: verbosity_arg = 0, minify: minify_arg = False, bcp: bcp_arg = False) -> None:
    transpile(input_path, verbosity, minify, bcp)
@app.command('r')
@app.command('run')
def run_cmd(input_path: pathlib.Path, verbosity: verbosity_arg = 0, minify: minify_arg = False, bcp: bcp_arg = False) -> None:
    input_path = Path(input_path)
    transpile(input_path, verbosity, minify, bcp)
    if input_path.is_dir():
        os.system(f'{sys.executable} -m dist')
        Path('dist').rmtree()
    else:
        os.system(f'{sys.executable} {input_path.with_suffix('.py')}')
        input_path.with_suffix('.py').remove()
    


app()